require 'apartment/elevators/generic'

begin
  require "apartment-sidekiq"
rescue LoadError
end

Apartment.configure do |config|
  config.excluded_models ||= []
  config.excluded_models |= ['PandaPal::Organization', 'PandaPal::Session']

  config.tenant_names = lambda {
    PandaPal::Organization.pluck(:name)
  }
end

Rails.application.config.middleware.use Apartment::Elevators::Generic, lambda { |request|
  if match = request.path.match(/\/(?:orgs?|organizations?)\/(\d+)/)
    PandaPal::Organization.find_by(id: match[1]).try(:name)
  elsif match = request.path.match(/\/(?:orgs?|organizations?|o)\/(\w+)/)
    PandaPal::Organization.find_by(name: match[1]).try(:name)
  elsif request.path.start_with?('/rails/active_storage/blobs/')
    PandaPal::Organization.find_by(id: request.params['organization_id']).try(:name)
  end
}

module PandaPal::Plugins::ApartmentCache
  private

  if Rails.version >= '5.0'
    def normalize_key(key, options)
      "tenant:#{Apartment::Tenant.current}/#{super}"
    end
  else
    def namespaced_key(*args, **kwargs)
      "tenant:#{Apartment::Tenant.current}/#{super}"
    end
  end
end
ActiveSupport::Cache::Store.send(:prepend, PandaPal::Plugins::ApartmentCache)

if defined?(ActionCable)
  module ActionCable
    module Channel
      class Base
        def self.broadcasting_for(model)
          # Rails 5 #stream_for passes #channel_name as part of model. Rails 6 doesn't and includes it via #broadcasting_for.
          model = [channel_name, model] unless model.is_a?(Array)
          serialize_broadcasting([ Apartment::Tenant.current, model ])
        end

        def self.serialize_broadcasting(object)
          case
          when object.is_a?(Array)
            object.map { |m| serialize_broadcasting(m) }.join(":")
          when object.respond_to?(:to_gid_param)
            object.to_gid_param
          else
            object.to_param
          end
        end
      end
    end
  end

  module PandaPal::Plugins::ActionCableApartment
    module Connection
      def tenant=(name)
        @tenant = name
      end

      def tenant
        @tenant || 'public'
      end

      def dispatch_websocket_message(*args, **kwargs)
        Apartment::Tenant.switch(tenant) do
          super
        end
      end
    end
  end

  ActionCable::Connection::Base.prepend(PandaPal::Plugins::ActionCableApartment::Connection)
end

if defined?(Delayed)
  module PandaPal::Plugins
    class ApartmentDelayedJobsPlugin < ::Delayed::Plugin
      callbacks do |lifecycle|
        lifecycle.around(:enqueue) do |job, *args, &block|
          current_tenant = Apartment::Tenant.current

          #make sure enqueue on public tenant unless we are testing since delayed job is set to run immediately
          Apartment::Tenant.switch!('public') unless Rails.env.test?
          job.tenant = current_tenant
          begin
            block.call(job, *args)
          rescue Exception => e
            Rails.logger.error("Error enqueing job #{job.to_s} - #{e.backtrace}")
          ensure
            #switch back to prev tenant
            Apartment::Tenant.switch!(current_tenant)
          end
        end

        lifecycle.before(:perform) do |worker, *args, &block|
          tenant = args.first.tenant
          Apartment::Tenant.switch!(tenant) if tenant.present?
          Rails.logger.debug("Running job with tenant #{Apartment::Tenant.current}")
        end

        lifecycle.around(:invoke_job) do |job, *args, &block|
          begin
            block.call(job, *args)
          ensure
            Apartment::Tenant.switch!('public')
            Rails.logger.debug("Resetting Tenant back to: #{Apartment::Tenant.current}")
          end
        end

        lifecycle.after(:failure) do |job, *args|
          Rails.logger.error("Job failed on tenant: #{Apartment::Tenant.current}")
        end
      end
    end
  end

  Delayed::Worker.plugins << PandaPal::Plugins::ApartmentDelayedJobsPlugin
end
